@@ -1,6 +1,6 @@
 
 /*
- * $Id: cache_cf.c,v 1.449 2006/06/30 21:23:04 hno Exp $
+ * $Id: cache_cf.c,v 1.480.2.7 2008/05/04 23:23:13 hno Exp $
  *
  * DEBUG: section 3     Configuration File Parsing
  * AUTHOR: Harvest Derived
@@ -34,6 +34,9 @@
  */
 
 #include "squid.h"
+#if HAVE_GLOB_H
+#include <glob.h>
+#endif
 
 #if SQUID_SNMP
 #include "snmp.h"
@@ -58,6 +61,8 @@ static const char *const list_sep = ", \t\n\r";
 
 static void parse_cachedir_option_readonly(SwapDir * sd, const char *option, const char *value, int reconfiguring);
 static void dump_cachedir_option_readonly(StoreEntry * e, const char *option, SwapDir * sd);
+static void parse_cachedir_option_minsize(SwapDir * sd, const char *option, const char *value, int reconfiguring);
+static void dump_cachedir_option_minsize(StoreEntry * e, const char *option, SwapDir * sd);
 static void parse_cachedir_option_maxsize(SwapDir * sd, const char *option, const char *value, int reconfiguring);
 static void dump_cachedir_option_maxsize(StoreEntry * e, const char *option, SwapDir * sd);
 static void parse_logformat(logformat ** logformat_definitions);
@@ -66,11 +71,16 @@ static void dump_logformat(StoreEntry * entry, const char *name, logformat * def
 static void dump_access_log(StoreEntry * entry, const char *name, customlog * definitions);
 static void free_logformat(logformat ** definitions);
 static void free_access_log(customlog ** definitions);
+static void parse_zph_mode(enum zph_mode *mode);
+static void dump_zph_mode(StoreEntry * entry, const char *name, enum zph_mode mode);
+static void free_zph_mode(enum zph_mode *mode);
 
 
 static struct cache_dir_option common_cachedir_options[] =
 {
-    {"read-only", parse_cachedir_option_readonly, dump_cachedir_option_readonly},
+    {"no-store", parse_cachedir_option_readonly, dump_cachedir_option_readonly},
+    {"read-only", parse_cachedir_option_readonly, NULL},
+    {"min-size", parse_cachedir_option_minsize, dump_cachedir_option_minsize},
     {"max-size", parse_cachedir_option_maxsize, dump_cachedir_option_maxsize},
     {NULL, NULL}
 };
@@ -124,6 +134,11 @@ static void free_https_port_list(https_port_list **);
 static int check_null_https_port_list(const https_port_list *);
 #endif
 #endif /* USE_SSL */
+static void parse_programline(wordlist **);
+static void free_programline(wordlist **);
+static void dump_programline(StoreEntry *, const char *, const wordlist *);
+
+static int parseOneConfigFile(const char *file_name, int depth);
 
 void
 self_destruct(void)
@@ -228,17 +243,26 @@ xatol(const char *token)
 {
     char *end;
     long ret = strtol(token, &end, 10);
-    if (ret == 0 && end == token)
+    if (end == token || *end)
 	self_destruct();
     return ret;
 }
 
-static int
+int
 xatoi(const char *token)
 {
     return xatol(token);
 }
 
+unsigned short
+xatos(const char *token)
+{
+    long port = xatol(token);
+    if (port & ~0xFFFF)
+	self_destruct();
+    return port;
+}
+
 static double
 xatof(const char *token)
 {
@@ -265,6 +289,15 @@ GetInteger(void)
     return i;
 }
 
+static u_short
+GetShort(void)
+{
+    char *token = strtok(NULL, w_space);
+    if (token == NULL)
+	self_destruct();
+    return xatos(token);
+}
+
 static squid_off_t
 GetOffT(void)
 {
@@ -299,24 +332,62 @@ update_maxobjsize(void)
     store_maxobjsize = ms;
 }
 
-int
-parseConfigFile(const char *file_name)
+static int
+parseManyConfigFiles(char *files, int depth)
+{
+    int error_count = 0;
+    char *saveptr = NULL;
+#if HAVE_GLOB
+    char *path;
+    glob_t globbuf;
+    int i;
+    memset(&globbuf, 0, sizeof(globbuf));
+    for (path = strwordtok(files, &saveptr); path; path = strwordtok(NULL, &saveptr)) {
+	if (glob(path, globbuf.gl_pathc ? GLOB_APPEND : 0, NULL, &globbuf) != 0) {
+	    fatalf("Unable to find configuration file: %s: %s",
+		path, xstrerror());
+	}
+    }
+    for (i = 0; i < globbuf.gl_pathc; i++) {
+	error_count += parseOneConfigFile(globbuf.gl_pathv[i], depth);
+    }
+    globfree(&globbuf);
+#else
+    char *file = strwordtok(files, &saveptr);
+    while (file != NULL) {
+	error_count += parseOneConfigFile(file, depth);
+	file = strwordtok(NULL, &saveptr);
+    }
+#endif /* HAVE_GLOB */
+    return error_count;
+}
+
+static int
+parseOneConfigFile(const char *file_name, int depth)
 {
     FILE *fp = NULL;
+    const char *orig_cfg_filename = cfg_filename;
+    int orig_config_lineno = config_lineno;
     char *token = NULL;
     char *tmp_line = NULL;
     int tmp_line_len = 0;
     size_t config_input_line_len;
     int err_count = 0;
-    configFreeMemory();
-    default_all();
+
+    debug(3, 1) ("Including Configuration File: %s (depth %d)\n", file_name, depth);
+    if (depth > 16) {
+	debug(0, 0) ("WARNING: can't include %s: includes are nested too deeply (>16)!\n", file_name);
+	return 1;
+    }
     if ((fp = fopen(file_name, "r")) == NULL)
 	fatalf("Unable to open configuration file: %s: %s",
 	    file_name, xstrerror());
 #ifdef _SQUID_WIN32_
     setmode(fileno(fp), O_TEXT);
 #endif
+
     cfg_filename = file_name;
+
     if ((token = strrchr(cfg_filename, '/')))
 	cfg_filename = token + 1;
     memset(config_input_line, '\0', BUFSIZ);
@@ -338,30 +409,51 @@ parseConfigFile(const char *file_name)
 	tmp_line_len += config_input_line_len;
 
 	if (tmp_line[tmp_line_len - 1] == '\\') {
-	    debug(3, 5) ("parseConfigFile: tmp_line='%s'\n", tmp_line);
+	    debug(3, 5) ("parseOneConfigFile: tmp_line='%s'\n", tmp_line);
 	    tmp_line[--tmp_line_len] = '\0';
 	    continue;
 	}
 	debug(3, 5) ("Processing: '%s'\n", tmp_line);
-	if (!parse_line(tmp_line)) {
-	    debug(3, 0) ("parseConfigFile: line %d unrecognized: '%s'\n",
+
+	/* Handle includes here */
+	if (tmp_line_len >= 9 &&
+	    strncmp(tmp_line, "include", 7) == 0 &&
+	    xisspace(tmp_line[7])) {
+	    err_count += parseManyConfigFiles(tmp_line + 8, depth + 1);
+	} else if (!parse_line(tmp_line)) {
+	    debug(3, 0) ("parseConfigFile: %s:%d unrecognized: '%s'\n",
+		cfg_filename,
 		config_lineno,
-		config_input_line);
+		tmp_line);
 	    err_count++;
 	}
 	safe_free(tmp_line);
 	tmp_line_len = 0;
     }
     fclose(fp);
+    cfg_filename = orig_cfg_filename;
+    config_lineno = orig_config_lineno;
+    return err_count;
+}
+
+int
+parseConfigFile(const char *file_name)
+{
+    int ret;
+
+    configFreeMemory();
+    default_all();
+    ret = parseOneConfigFile(file_name, 0);
     defaults_if_none();
     configDoConfigure();
+
     if (opt_send_signal == -1) {
 	cachemgrRegister("config",
 	    "Current Squid Configuration",
 	    dump_config,
 	    1, 1);
     }
-    return err_count;
+    return ret;
 }
 
 static void
@@ -445,12 +537,16 @@ configDoConfigure(void)
 #if USE_UNLINKD
     requirePathnameExists("unlinkd_program", Config.Program.unlinkd);
 #endif
+    requirePathnameExists("logfile_daemon", Config.Program.logfile_daemon);
     if (Config.Program.url_rewrite.command)
 	requirePathnameExists("url_rewrite_program", Config.Program.url_rewrite.command->key);
     if (Config.Program.location_rewrite.command)
 	requirePathnameExists("location_rewrite_program", Config.Program.location_rewrite.command->key);
     requirePathnameExists("Icon Directory", Config.icons.directory);
     requirePathnameExists("Error Directory", Config.errorDirectory);
+    authenticateConfigure(&Config.authConfig);
+    externalAclConfigure();
+    refreshCheckConfigure();
 #if HTTP_VIOLATIONS
     {
 	const refresh_t *R;
@@ -466,6 +562,12 @@ configDoConfigure(void)
 	    debug(22, 1) ("WARNING: use of 'override-lastmod' in 'refresh_pattern' violates HTTP\n");
 	    break;
 	}
+	for (R = Config.Refresh; R; R = R->next) {
+	    if (R->stale_while_revalidate <= 0)
+		continue;
+	    debug(22, 1) ("WARNING: use of 'stale-while-revalidate' in 'refresh_pattern' violates HTTP\n");
+	    break;
+	}
     }
 #endif
 #if !HTTP_VIOLATIONS
@@ -474,13 +576,6 @@ configDoConfigure(void)
     if (!Config.onoff.via)
 	debug(22, 1) ("WARNING: HTTP requires the use of Via\n");
 #endif
-    if (Config.Wais.relayHost) {
-	if (Config.Wais.peer)
-	    cbdataFree(Config.Wais.peer);
-	Config.Wais.peer = cbdataAlloc(peer);
-	Config.Wais.peer->host = xstrdup(Config.Wais.relayHost);
-	Config.Wais.peer->http_port = Config.Wais.relayPort;
-    }
     if (aclPurgeMethodInUse(Config.accessList.http))
 	Config2.onoff.enable_purge = 1;
     if (geteuid() == 0) {
@@ -662,21 +757,25 @@ parseBytesUnits(const char *unit)
 static void
 dump_acl(StoreEntry * entry, const char *name, acl * ae)
 {
-    wordlist *w;
-    wordlist *v;
     while (ae != NULL) {
 	debug(3, 3) ("dump_acl: %s %s\n", name, ae->name);
-	v = w = aclDumpGeneric(ae);
-	while (v != NULL) {
-	    debug(3, 3) ("dump_acl: %s %s %s\n", name, ae->name, v->key);
-	    storeAppendPrintf(entry, "%s %s %s %s\n",
-		name,
-		ae->name,
-		aclTypeToStr(ae->type),
-		v->key);
-	    v = v->next;
+	if (strstr(ae->cfgline, " \""))
+	    storeAppendPrintf(entry, "%s\n", ae->cfgline);
+	else {
+	    wordlist *w;
+	    wordlist *v;
+	    v = w = aclDumpGeneric(ae);
+	    while (v != NULL) {
+		debug(3, 3) ("dump_acl: %s %s %s\n", name, ae->name, v->key);
+		storeAppendPrintf(entry, "%s %s %s %s\n",
+		    name,
+		    ae->name,
+		    aclTypeToStr(ae->type),
+		    v->key);
+		v = v->next;
+	    }
+	    wordlistDestroy(&w);
 	}
-	wordlistDestroy(&w);
 	ae = ae->next;
     }
 }
@@ -1025,7 +1124,7 @@ parse_delay_pool_access(delayConfig * cfg)
 
     parse_ushort(&pool);
     if (pool < 1 || pool > cfg->pools) {
-	debug(3, 0) ("parse_delay_pool_rates: Ignoring pool %d not in 1 .. %d\n", pool, cfg->pools);
+	debug(3, 0) ("parse_delay_pool_access: Ignoring pool %d not in 1 .. %d\n", pool, cfg->pools);
 	return;
     }
     aclParseAccessLine(&cfg->access[pool - 1]);
@@ -1216,7 +1315,8 @@ dump_cachedir_options(StoreEntry * entry, struct cache_dir_option *options, Swap
     if (!options)
 	return;
     for (option = options; option->name; option++)
-	option->dump(entry, option->name, sd);
+	if (option->dump)
+	    option->dump(entry, option->name, sd);
 }
 
 static void
@@ -1395,6 +1495,7 @@ parse_cachedir(cacheSwap * swap)
     sd = swap->swapDirs + swap->n_configured;
     sd->type = storefs_list[fs].typestr;
     /* defaults in case fs implementation fails to set these */
+    sd->min_objsize = 0;
     sd->max_objsize = -1;
     sd->fs.blksize = 1024;
     /* parse the FS parameters and options */
@@ -1422,6 +1523,29 @@ dump_cachedir_option_readonly(StoreEntry * e, const char *option, SwapDir * sd)
 	storeAppendPrintf(e, " %s", option);
 }
 
+static void
+parse_cachedir_option_minsize(SwapDir * sd, const char *option, const char *value, int reconfiguring)
+{
+    squid_off_t size;
+
+    if (!value)
+	self_destruct();
+
+    size = strto_off_t(value, NULL, 10);
+
+    if (reconfiguring && sd->min_objsize != size)
+	debug(3, 1) ("Cache dir '%s' min object size now %ld\n", sd->path, (long int) size);
+
+    sd->min_objsize = size;
+}
+
+static void
+dump_cachedir_option_minsize(StoreEntry * e, const char *option, SwapDir * sd)
+{
+    if (sd->min_objsize != 0)
+	storeAppendPrintf(e, " %s=%ld", option, (long int) sd->min_objsize);
+}
+
 static void
 parse_cachedir_option_maxsize(SwapDir * sd, const char *option, const char *value, int reconfiguring)
 {
@@ -1483,7 +1607,7 @@ parse_cachedir_options(SwapDir * sd, struct cache_dir_option *options, int recon
     if (reconfiguring) {
 	if (old_read_only != sd->flags.read_only) {
 	    debug(3, 1) ("Cache dir '%s' now %s\n",
-		sd->path, sd->flags.read_only ? "Read-Only" : "Read-Write");
+		sd->path, sd->flags.read_only ? "No-Store" : "Read-Write");
 	}
     }
 }
@@ -1560,18 +1684,47 @@ dump_peer(StoreEntry * entry, const char *name, peer * p)
     }
 }
 
+static u_short
+GetService(const char *proto)
+{
+    struct servent *port = NULL;
+    char *token = strtok(NULL, w_space);
+    if (token == NULL) {
+	self_destruct();
+	return -1;		/* NEVER REACHED */
+    }
+    port = getservbyname(token, proto);
+    if (port != NULL) {
+	return ntohs((u_short) port->s_port);
+    }
+    return xatos(token);
+}
+
+static u_short
+GetTcpService(void)
+{
+    return GetService("tcp");
+}
+
+static u_short
+GetUdpService(void)
+{
+    return GetService("udp");
+}
+
 static void
 parse_peer(peer ** head)
 {
     char *token = NULL;
     peer *p;
-    int i;
     p = cbdataAlloc(peer);
     p->http_port = CACHE_HTTP_PORT;
     p->icp.port = CACHE_ICP_PORT;
     p->weight = 1;
     p->stats.logged_state = PEER_ALIVE;
     p->monitor.state = PEER_ALIVE;
+    p->monitor.interval = 300;
+    p->tcp_up = PEER_TCP_MAGIC_COUNT;
     if ((token = strtok(NULL, w_space)) == NULL)
 	self_destruct();
     p->host = xstrdup(token);
@@ -1579,10 +1732,14 @@ parse_peer(peer ** head)
     if ((token = strtok(NULL, w_space)) == NULL)
 	self_destruct();
     p->type = parseNeighborType(token);
-    i = GetInteger();
-    p->http_port = (u_short) i;
-    i = GetInteger();
-    p->icp.port = (u_short) i;
+    if (p->type == PEER_MULTICAST) {
+	p->options.no_digest = 1;
+	p->options.no_netdb_exchange = 1;
+    }
+    p->http_port = GetTcpService();
+    if (!p->http_port)
+	self_destruct();
+    p->icp.port = GetUdpService();
     p->connection_auth = -1;	/* auto */
     while ((token = strtok(NULL, w_space))) {
 	if (!strcasecmp(token, "proxy-only")) {
@@ -1593,6 +1750,10 @@ parse_peer(peer ** head)
 	    p->options.no_digest = 1;
 	} else if (!strcasecmp(token, "multicast-responder")) {
 	    p->options.mcast_responder = 1;
+#if PEER_MULTICAST_SIBLINGS
+	} else if (!strcasecmp(token, "multicast-siblings")) {
+	    p->options.mcast_siblings = 1;
+#endif
 	} else if (!strncasecmp(token, "weight=", 7)) {
 	    p->weight = xatoi(token + 7);
 	} else if (!strcasecmp(token, "closest-only")) {
@@ -1650,18 +1811,23 @@ parse_peer(peer ** head)
 	    if (token[5])
 		p->name = xstrdup(token + 5);
 	} else if (!strncasecmp(token, "monitorurl=", 11)) {
+	    char *url = token + 11;
 	    safe_free(p->monitor.url);
-	    if (token[11])
-		p->monitor.url = xstrdup(token + 11);
+	    if (*url == '/') {
+		int size = strlen("http://") + strlen(p->host) + 16 + strlen(url);
+		p->monitor.url = xmalloc(size);
+		snprintf(p->monitor.url, size, "http://%s:%d%s", p->host, p->http_port, url);
+	    } else {
+		p->monitor.url = xstrdup(url);
+	    }
 	} else if (!strncasecmp(token, "monitorsize=", 12)) {
+	    char *token2 = strchr(token + 12, ',');
+	    if (!token2)
+		token2 = strchr(token + 12, '-');
+	    if (token2)
+		*token2++ = '\0';
 	    p->monitor.min = xatoi(token + 12);
-	    p->monitor.max = -1;
-	    if (strchr(token + 12, ','))
-		token = strchr(token + 12, ',');
-	    else
-		token = strchr(token + 12, '-');
-	    if (token)
-		p->monitor.max = xatoi(token + 1);
+	    p->monitor.max = token2 ? xatoi(token2) : -1;
 	} else if (!strncasecmp(token, "monitorinterval=", 16)) {
 	    p->monitor.interval = xatoi(token + 16);
 	} else if (!strncasecmp(token, "monitortimeout=", 15)) {
@@ -1719,6 +1885,10 @@ parse_peer(peer ** head)
 	    p->connection_auth = 1;
 	} else if (strcmp(token, "connection-auth=auto") == 0) {
 	    p->connection_auth = -1;
+	} else if (strncmp(token, "idle=", 5) == 0) {
+	    p->idle = xatoi(token + 5);
+	} else if (strcmp(token, "http11") == 0) {
+	    p->options.http11 = 1;
 	} else {
 	    debug(3, 0) ("parse_peer: token='%s'\n", token);
 	    self_destruct();
@@ -1729,7 +1899,6 @@ parse_peer(peer ** head)
     if (p->weight < 1)
 	p->weight = 1;
     p->icp.version = ICP_VERSION_CURRENT;
-    p->tcp_up = PEER_TCP_MAGIC_COUNT;
     p->test_fd = -1;
 #if USE_CACHE_DIGESTS
     if (!p->options.no_digest) {
@@ -1961,16 +2130,13 @@ static void
 parse_ushortlist(ushortlist ** P)
 {
     char *token;
-    int i;
+    u_short i;
     ushortlist *u;
     ushortlist **U;
     while ((token = strtok(NULL, w_space))) {
-	if (sscanf(token, "%d", &i) != 1)
-	    self_destruct();
-	if (i < 0)
-	    i = 0;
+	i = GetShort();
 	u = xcalloc(1, sizeof(ushortlist));
-	u->i = (u_short) i;
+	u->i = i;
 	for (U = P; *U; U = &(*U)->next);
 	*U = u;
     }
@@ -2084,7 +2250,15 @@ dump_refreshpattern(StoreEntry * entry, const char *name, refresh_t * head)
 	    storeAppendPrintf(entry, " ignore-private");
 	if (head->flags.ignore_auth)
 	    storeAppendPrintf(entry, " ignore-auth");
+	if (head->stale_while_revalidate > 0)
+	    storeAppendPrintf(entry, " stale-while-revalidate=%d", head->stale_while_revalidate);
 #endif
+	if (head->flags.ignore_stale_while_revalidate)
+	    storeAppendPrintf(entry, " ignore-stale-while-revalidate");
+	if (head->max_stale >= 0)
+	    storeAppendPrintf(entry, " max-stale=%d", head->max_stale);
+	if (head->negative_ttl >= 0)
+	    storeAppendPrintf(entry, " negative-ttl=%d", head->negative_ttl);
 	storeAppendPrintf(entry, "\n");
 	head = head->next;
     }
@@ -2107,6 +2281,10 @@ parse_refreshpattern(refresh_t ** head)
     int ignore_private = 0;
     int ignore_auth = 0;
 #endif
+    int stale_while_revalidate = -1;
+    int ignore_stale_while_revalidate = 0;
+    int max_stale = -1;
+    int negative_ttl = -1;
     int i;
     refresh_t *t;
     regex_t comp;
@@ -2151,17 +2329,27 @@ parse_refreshpattern(refresh_t ** head)
 	    ignore_reload = 1;
 	    refresh_nocache_hack = 1;
 	    /* tell client_side.c that this is used */
+	} else if (!strncmp(token, "stale-while-revalidate=", 23)) {
+	    stale_while_revalidate = atoi(token + 23);
 	} else
 #endif
-	    debug(22, 0) ("redreshAddToList: Unknown option '%s': %s\n",
+	if (!strncmp(token, "max-stale=", 10)) {
+	    max_stale = atoi(token + 10);
+	} else if (!strncmp(token, "negative-ttl=", 13)) {
+	    negative_ttl = atoi(token + 13);
+	} else if (!strcmp(token, "ignore-stale-while-revalidate")) {
+	    ignore_stale_while_revalidate = 1;
+	} else {
+	    debug(22, 0) ("parse_refreshpattern: Unknown option '%s': %s\n",
 		pattern, token);
+	}
     }
     if ((errcode = regcomp(&comp, pattern, flags)) != 0) {
 	char errbuf[256];
 	regerror(errcode, &comp, errbuf, sizeof errbuf);
 	debug(22, 0) ("%s line %d: %s\n",
 	    cfg_filename, config_lineno, config_input_line);
-	debug(22, 0) ("refreshAddToList: Invalid regular expression '%s': %s\n",
+	debug(22, 0) ("parse_refreshpattern: Invalid regular expression '%s': %s\n",
 	    pattern, errbuf);
 	return;
     }
@@ -2191,6 +2379,10 @@ parse_refreshpattern(refresh_t ** head)
     if (ignore_auth)
 	t->flags.ignore_auth = 1;
 #endif
+    t->flags.ignore_stale_while_revalidate = ignore_stale_while_revalidate;
+    t->stale_while_revalidate = stale_while_revalidate;
+    t->max_stale = max_stale;
+    t->negative_ttl = negative_ttl;
     t->next = NULL;
     while (*head)
 	head = &(*head)->next;
@@ -2248,7 +2440,7 @@ parse_eol(char *volatile *var)
     safe_free(*var);
     if (token == NULL)
 	self_destruct();
-    while (*token && isspace(*token))
+    while (*token && xisspace(*token))
 	token++;
     if (!*token)
 	self_destruct();
@@ -2395,16 +2587,11 @@ free_ushort(u_short * u)
 static void
 parse_ushort(u_short * var)
 {
-    int i;
-
-    i = GetInteger();
-    if (i < 0)
-	i = 0;
-    *var = (u_short) i;
+    *var = GetShort();
 }
 
 static void
-dump_wordlist(StoreEntry * entry, const char *name, wordlist * list)
+dump_wordlist(StoreEntry * entry, const char *name, const wordlist * list)
 {
     while (list != NULL) {
 	storeAppendPrintf(entry, "%s %s\n", name, list->key);
@@ -2482,8 +2669,7 @@ free_removalpolicy(RemovalPolicySettings ** settings)
 	return;
     free_string(&(*settings)->type);
     free_wordlist(&(*settings)->args);
-    xfree(*settings);
-    *settings = NULL;
+    safe_free(*settings);
 }
 
 static void
@@ -2590,37 +2776,10 @@ static void
 parse_sockaddr_in_list(sockaddr_in_list ** head)
 {
     char *token;
-    char *t;
-    char *host;
-    char *tmp;
-    const struct hostent *hp;
-    unsigned short port = 0;
     sockaddr_in_list *s;
     while ((token = strtok(NULL, w_space))) {
-	host = NULL;
-	port = 0;
-	if ((t = strchr(token, ':'))) {
-	    /* host:port */
-	    host = token;
-	    *t = '\0';
-	    port = (unsigned short) xatoi(t + 1);
-	    if (0 == port)
-		self_destruct();
-	} else if ((port = strtol(token, &tmp, 10)), !*tmp) {
-	    /* port */
-	} else {
-	    host = token;
-	    port = 0;
-	}
 	s = xcalloc(1, sizeof(*s));
-	s->s.sin_port = htons(port);
-	if (NULL == host)
-	    s->s.sin_addr = any_addr;
-	else if (1 == safe_inet_addr(host, &s->s.sin_addr))
-	    (void) 0;
-	else if ((hp = gethostbyname(host)))	/* dont use ipcache */
-	    s->s.sin_addr = inaddrFromHostent(hp);
-	else
+	if (!parse_sockaddr(token, &s->s))
 	    self_destruct();
 	while (*head)
 	    head = &(*head)->next;
@@ -2666,18 +2825,18 @@ parse_http_port_specification(http_port_list * s, char *token)
     const struct hostent *hp;
     unsigned short port = 0;
     char *t;
+    s->name = xstrdup(token);
     if ((t = strchr(token, ':'))) {
 	/* host:port */
 	host = token;
 	*t = '\0';
-	port = (unsigned short) xatoi(t + 1);
-	if (0 == port)
-	    self_destruct();
-    } else if ((port = xatoi(token)) > 0) {
-	/* port */
+	port = xatos(t + 1);
     } else {
-	self_destruct();
+	/* port */
+	port = xatos(token);
     }
+    if (port == 0)
+	self_destruct();
     s->s.sin_port = htons(port);
     if (NULL == host)
 	s->s.sin_addr = any_addr;
@@ -2710,7 +2869,9 @@ parse_http_port_option(http_port_list * s, char *token)
 	s->vport = ntohs(s->s.sin_port);
 	s->accel = 1;
     } else if (strncmp(token, "vport=", 6) == 0) {
-	s->vport = xatoi(token + 6);
+	s->vport = xatos(token + 6);
+	s->accel = 1;
+    } else if (strcmp(token, "accel") == 0) {
 	s->accel = 1;
     } else if (strcmp(token, "no-connection-auth") == 0) {
 	s->no_connection_auth = 1;
@@ -2723,11 +2884,44 @@ parse_http_port_option(http_port_list * s, char *token)
 	s->tproxy = 1;
 	need_linux_tproxy = 1;
 #endif
+    } else if (strcmp(token, "act-as-origin") == 0) {
+	s->act_as_origin = 1;
+	s->accel = 1;
+    } else if (strcmp(token, "allow-direct") == 0) {
+	s->allow_direct = 1;
+    } else if (strcmp(token, "http11") == 0) {
+	s->http11 = 1;
+    } else if (strcmp(token, "tcpkeepalive") == 0) {
+	s->tcp_keepalive.enabled = 1;
+    } else if (strncmp(token, "tcpkeepalive=", 13) == 0) {
+	char *t = token + 13;
+	s->tcp_keepalive.enabled = 1;
+	s->tcp_keepalive.idle = atoi(t);
+	t = strchr(t, ',');
+	if (t) {
+	    t++;
+	    s->tcp_keepalive.interval = atoi(t);
+	    t = strchr(t, ',');
+	}
+	if (t) {
+	    t++;
+	    s->tcp_keepalive.timeout = atoi(t);
+	    t = strchr(t, ',');
+	}
     } else {
 	self_destruct();
     }
 }
 
+static void
+verify_http_port_options(http_port_list * s)
+{
+    if (s->accel && s->transparent) {
+	debug(28, 0) ("Can't be both a transparent proxy and web server accelerator on the same port\n");
+	self_destruct();
+    }
+}
+
 static void
 free_generic_http_port_data(http_port_list * s)
 {
@@ -2760,6 +2954,7 @@ parse_http_port_list(http_port_list ** head)
     while ((token = strtok(NULL, w_space))) {
 	parse_http_port_option(s, token);
     }
+    verify_http_port_options(s);
     while (*head)
 	head = &(*head)->next;
     *head = s;
@@ -2772,16 +2967,37 @@ dump_generic_http_port(StoreEntry * e, const char *n, const http_port_list * s)
 	n,
 	inet_ntoa(s->s.sin_addr),
 	ntohs(s->s.sin_port));
-    if (s->defaultsite)
-	storeAppendPrintf(e, " defaultsite=%s", s->defaultsite);
     if (s->transparent)
 	storeAppendPrintf(e, " transparent");
+    if (s->accel)
+	storeAppendPrintf(e, " accel");
+    if (s->defaultsite)
+	storeAppendPrintf(e, " defaultsite=%s", s->defaultsite);
     if (s->vhost)
 	storeAppendPrintf(e, " vhost");
-    if (s->vport)
+    if (s->vport == -1)
 	storeAppendPrintf(e, " vport");
+    else if (s->vport)
+	storeAppendPrintf(e, " vport=%d", s->vport);
+    if (s->urlgroup)
+	storeAppendPrintf(e, " urlgroup=%s", s->urlgroup);
+    if (s->protocol)
+	storeAppendPrintf(e, " protocol=%s", s->protocol);
     if (s->no_connection_auth)
 	storeAppendPrintf(e, " no-connection-auth");
+#if LINUX_TPROXY
+    if (s->tproxy)
+	storeAppendPrintf(e, " tproxy");
+#endif
+    if (s->http11)
+	storeAppendPrintf(e, " http11");
+    if (s->tcp_keepalive.enabled) {
+	if (s->tcp_keepalive.idle || s->tcp_keepalive.interval || s->tcp_keepalive.timeout) {
+	    storeAppendPrintf(e, " tcp_keepalive=%d,%d,%d", s->tcp_keepalive.idle, s->tcp_keepalive.interval, s->tcp_keepalive.timeout);
+	} else {
+	    storeAppendPrintf(e, " tcp_keepalive");
+	}
+    }
 }
 static void
 dump_http_port_list(StoreEntry * e, const char *n, const http_port_list * s)
@@ -2880,6 +3096,7 @@ parse_https_port_list(https_port_list ** head)
 	    parse_http_port_option(&s->http, token);
 	}
     }
+    verify_http_port_options(&s->http);
     while (*head)
 	head = (https_port_list **) (void *) (&(*head)->http.next);
     s->sslContext = sslCreateServerContext(s->cert, s->key, s->version, s->cipher, s->options, s->sslflags, s->clientca, s->cafile, s->capath, s->crlfile, s->dhfile, s->sslcontext);
@@ -2962,7 +3179,7 @@ requirePathnameExists(const char *name, const char *path)
 	path = pathbuf;
     }
     if (stat(path, &sb) < 0) {
-	if (opt_send_signal == -1 || opt_send_signal == SIGHUP)
+	if ((opt_send_signal == -1 || opt_send_signal == SIGHUP) && !opt_parse_cfg_only)
 	    fatalf("%s %s: %s", name, path, xstrerror());
 	else
 	    fprintf(stderr, "WARNING: %s %s: %s\n", name, path, xstrerror());
@@ -3170,3 +3387,69 @@ free_access_log(customlog ** definitions)
 	xfree(log);
     }
 }
+
+static void
+parse_programline(wordlist ** line)
+{
+    if (*line)
+	self_destruct();
+    parse_wordlist(line);
+}
+
+static void
+free_programline(wordlist ** line)
+{
+    free_wordlist(line);
+}
+
+static void
+dump_programline(StoreEntry * entry, const char *name, const wordlist * line)
+{
+    dump_wordlist(entry, name, line);
+}
+
+static void
+parse_zph_mode(enum zph_mode *mode)
+{
+    char *token = strtok(NULL, w_space);
+    if (!token)
+	self_destruct();
+    if (strcmp(token, "off") == 0)
+	*mode = ZPH_OFF;
+    else if (strcmp(token, "tos") == 0)
+	*mode = ZPH_TOS;
+    else if (strcmp(token, "priority") == 0)
+	*mode = ZPH_PRIORITY;
+    else if (strcmp(token, "option") == 0)
+	*mode = ZPH_OPTION;
+    else {
+	debug(3, 0) ("WARNING: unsupported zph_mode argument '%s'\n", token);
+    }
+}
+
+static void
+dump_zph_mode(StoreEntry * entry, const char *name, enum zph_mode mode)
+{
+    const char *modestr = "unknown";
+    switch (mode) {
+    case ZPH_OFF:
+	modestr = "off";
+	break;
+    case ZPH_TOS:
+	modestr = "tos";
+	break;
+    case ZPH_PRIORITY:
+	modestr = "priority";
+	break;
+    case ZPH_OPTION:
+	modestr = "option";
+	break;
+    }
+    storeAppendPrintf(entry, "%s %s\n", name, modestr);
+}
+
+static void
+free_zph_mode(enum zph_mode *mode)
+{
+    *mode = ZPH_OFF;
+}
